use core::f32;

use fasthash::murmur2::hash64_with_seed;

// rust needs to add hex literals. this is 0x1p-32
const ONE_OVER_MAX: f32 = 2.3283064e-10;

const SOBOL_MATRIX: [u32; 32] = [
    0x80000000, 0xc0000000, 0xa0000000, 0xf0000000, 0x88000000, 0xcc000000, 0xaa000000, 0xff000000,
    0x80800000, 0xc0c00000, 0xa0a00000, 0xf0f00000, 0x88880000, 0xcccc0000, 0xaaaa0000, 0xffff0000,
    0x80008000, 0xc000c000, 0xa000a000, 0xf000f000, 0x88008800, 0xcc00cc00, 0xaa00aa00, 0xff00ff00,
    0x80808080, 0xc0c0c0c0, 0xa0a0a0a0, 0xf0f0f0f0, 0x88888888, 0xcccccccc, 0xaaaaaaaa, 0xffffffff,
];

#[rustfmt::skip]
const PERMUTATIONS: [[u8; 4]; 24] = [
    [0, 1, 2, 3], [0, 1, 3, 2], [0, 2, 1, 3], [0, 2, 3, 1], [0, 3, 1, 2], [0, 3, 2, 1],
    [1, 0, 2, 3], [1, 0, 3, 2], [1, 2, 0, 3], [1, 2, 3, 0], [1, 3, 0, 2], [1, 3, 2, 0],
    [2, 0, 1, 3], [2, 0, 3, 1], [2, 1, 0, 3], [2, 1, 3, 0], [2, 3, 0, 1], [2, 3, 1, 0],
    [3, 0, 1, 2], [3, 0, 2, 1], [3, 1, 0, 2], [3, 1, 2, 0], [3, 2, 0, 1], [3, 2, 1, 0],
];

pub struct Sampler {
    seed: u32,
    samples: u32,
    log2_samples: u32,
    base4_digits: u32,
}

impl Sampler {
    pub fn new(width: u32, height: u32, samples: u32, seed: u32) -> Self {
        let res = u32::next_power_of_two(u32::max(width, height));
        let log2_samples = u32::ilog2(samples);
        let base4_digits = u32::ilog2(res) + log2_samples.div_ceil(2);

        Self {
            seed,
            samples,
            log2_samples,
            base4_digits,
        }
    }

    pub fn pixel(&mut self, x: u32, y: u32) -> impl Iterator<Item = PixelSample> {
        let &mut Self {
            seed,
            samples,
            log2_samples,
            base4_digits,
            ..
        } = self;

        (0..samples).map(move |i| {
            let morton_index = (morton_2d(x, y) << log2_samples) | i;

            PixelSample {
                seed,
                log2_samples,
                base4_digits,
                morton_index,
                dimension: 0,
            }
        })
    }
}

pub struct PixelSample {
    seed: u32,
    log2_samples: u32,
    base4_digits: u32,
    morton_index: u32,
    dimension: u32,
}

impl PixelSample {
    pub fn get_1d(&mut self) -> f32 {
        let index = self.get_sample_index();
        self.dimension += 1;

        let hash = hash64_with_seed(self.dimension.to_ne_bytes(), self.seed as u64);
        let seed = hash as u32;

        sobol_sample0(index, seed)
    }

    pub fn get_2d(&mut self) -> (f32, f32) {
        let index = self.get_sample_index();
        self.dimension += 2;

        let hash = hash64_with_seed(self.dimension.to_ne_bytes(), self.seed as u64);

        let lo = hash as u32;
        let hi = (hash >> 32) as u32;

        (sobol_sample0(index, lo), sobol_sample1(index, hi))
    }

    fn get_sample_index(&self) -> u32 {
        let mut index: u32 = 0;
        let last_digit = self.log2_samples & 1;

        for i in (last_digit..self.base4_digits).rev() {
            let digit_shift = 2 * i - last_digit;
            let digit = (self.morton_index >> digit_shift) & 3;
            let higher_bits = self.morton_index >> (digit_shift + 2);
            let mixed = mix_bits((higher_bits ^ (0x55555555 * self.dimension)) as u64);
            let p = (mixed >> 24) as usize % PERMUTATIONS.len();
            let digit = PERMUTATIONS[p][digit as usize];
            index |= (digit as u32) << digit_shift;
        }

        if last_digit != 0 {
            let digit = self.morton_index & 1;
            let higher_bits = self.morton_index >> 1;
            let mixed = mix_bits((higher_bits ^ (0x55555555 * self.dimension)) as u64) as u32;
            index |= (digit ^ mixed) & 1;
        }

        index
    }
}

fn sobol_sample0(a: u32, seed: u32) -> f32 {
    let v = a.reverse_bits();
    f32::min(fast_owen(v, seed) as f32 * ONE_OVER_MAX, 1.0 - f32::EPSILON)
}

fn sobol_sample1(mut a: u32, seed: u32) -> f32 {
    let mut v: u32 = 0;
    for m in SOBOL_MATRIX {
        if a & 1 != 0 {
            v ^= m;
        }
        a >>= 1;
        if a == 0 {
            break;
        }
    }
    f32::min(fast_owen(v, seed) as f32 * ONE_OVER_MAX, 1.0 - f32::EPSILON)
}

// magic. taken from pbrt.
#[inline(always)]
fn fast_owen(mut v: u32, seed: u32) -> u32 {
    v = v.reverse_bits();
    v ^= v.wrapping_mul(0x3d20adea);
    v = v.wrapping_add(seed);
    v = v.wrapping_mul((seed >> 16) | 1);
    v ^= v.wrapping_mul(0x05526c56);
    v ^= v.wrapping_mul(0x53a22864);
    v.reverse_bits()
}

// more magic. also taken from pbrt.
#[inline(always)]
fn mix_bits(mut v: u64) -> u64 {
    v ^= v >> 31;
    v = v.wrapping_mul(0x7fb5d329728ea185);
    v ^= v >> 27;
    v = v.wrapping_mul(0x81dadef4bc2dd44d);
    v ^ (v >> 33)
}

fn separate(mut v: u32) -> u32 {
    v &= 0x0000ffff;
    v = (v | (v << 8)) & 0x00ff00ff;
    v = (v | (v << 4)) & 0x0f0f0f0f;
    v = (v | (v << 2)) & 0x33333333;
    v = (v | (v << 1)) & 0x55555555;
    v
}

fn morton_2d(x: u32, y: u32) -> u32 {
    (separate(y) << 1) | separate(x)
}
